#include <proto.h>

#include "gtlist.h"
#include "gtapp.h"
#include "gterror.h"
#include "gtresname.h"
#include "gtcommon.h"

/************************************************************************\
*
* GtList class.
*
* $Ident: This file is part of the 'Gt-library' 1.0.0 $
* $Copyright: (C) 1997 by Gran Gustafsson $
*
\************************************************************************/

#define GTLF_PRESS		0x02

/************************************************************************\
*
* Create and init list.
*
\************************************************************************/
GtList *GtMkList(char *n,GtManager *p,GtTag *tgs)
{
	GtList *obj;
	
	if(!(obj=new GtList()))
	{
		GtError(NULL,GTEC_NOOBJECT,"No GtList object!\n");
		return(NULL);
	}
	if(!obj->Init(n,p,tgs))
	{
		GtDestroy(obj);
		obj=NULL;
	}
	return(obj);
}
	
/************************************************************************\
*
* Create and init list.
*
\************************************************************************/
GtList *GtMkList(char *n,GtManager *p,ulong tag,...)
{
	va_list va;
	GtList *obj;
	int ps;
	
	va_start(va,tag);
	ps=GtBeginVaTags();
	obj=GtMkList(n,p,GtCopyVaTags(tag,&va));
	GtEndVaTags(ps);
	va_end(va);
	return(obj);
}

/************************************************************************\
*
* Setup empty list.
*
\************************************************************************/
GtList::GtList(void)
{
	GtDInit(&ihd);
	topi=NULL;
	seli=NULL;
	curri=NULL;
	topnr=0;
	visnr=0;
	totnr=0;
	last=0;
	space=1;
	foff=0;
	fh=0;
	mclk=1;
	mode=GTLM_READ;
	lflags=0;
	ftype=GTF_LAST_NR+1;
	ibw=0;
	bdrw=1;
	ibdrw=2;
}

/************************************************************************\
*
* Destroy list.
*
\************************************************************************/
void GtList::Destroy(void)
{
	GtListItem *l;
	
	while((l=(GtListItem *)GtDRemHead(&ihd)))
		free(l);
	GtWidget::Destroy();
}

/************************************************************************\
*
* Setup list.
*
\************************************************************************/
sbool GtList::Init(char *n,GtManager *p,GtTag *tgs)
{
	GtTag *t;
	XCharStruct xcs;
	
	if(!GtWidget::Init(n,p,tgs))
		return(FALSE);
	app->GetRes(this,GTR_SPACE,GTRT_WORD,&space);
	app->GetRes(this,GTR_IBDR,GTRT_BYTE,&ibdrw);
	app->GetRes(this,GTR_BDR,GTRT_BYTE,&bdrw);
	app->GetRes(this,GTR_FRAME,GTRT_FRAME,&ftype);
	while((t=GtNextTag(&tgs)))
	{
		switch(t->tag)
		{
			case GT_Mode:
				mode=t->d.val;
				if(mode>GTLM_LAST_NR)
					mode=GTLM_READ;
				break;
			case GT_Spacing:
				space=t->d.val;
				if(space<0)
					space=0;
				break;
			case GT_FrameType:
				ftype=t->d.val;
				break;
			case GT_InnerBorder:
				ibdrw=t->d.val;
				break;
			case GT_Border:
				bdrw=t->d.val;
				break;
		}
	}
	if(!this->PrivateGC(GT_Foreground))
		return(FALSE);
	ibw=ibdrw+bdrw+(space ? 1 : 0);
	gtfnt->MaxExtents(&xcs);
	foff=xcs.ascent;
	fh=foff+xcs.descent;
	this->SetMinMaxSize(GTMAX(xcs.width+2*ibw,2*ibw+1),GTMAX(fh+2*ibw,2*ibw+1),
											GTS_MAX,GTS_MAX);
	return(TRUE);
}

/************************************************************************\
*
* Set list attributes.
*
\************************************************************************/
sbool GtList::Set(GtTag *tgs)
{
	GtTag *t;
	long tnr;
	ulong col;
	sbool vch=FALSE,mch=FALSE,tch=FALSE,fch=FALSE,cch=FALSE;
	XCharStruct xcs;
	GtListItem *l;
	GtListSelectCD cd;
	
	col=gfx->cols[GTFORE];
	if(!GtWidget::Set(tgs))
		return(FALSE);
	tnr=topnr;
	while((t=GtNextTag(&tgs)))
	{
		switch(t->tag)
		{
			case GT_Mode:
				mode=t->d.val;
				if(mode>GTLM_LAST_NR)
					mode=GTLM_READ;
				mch=TRUE;
				break;
			case GT_Spacing:
				space=t->d.val;
				if(space<0)
					space=0;
				vch=TRUE;
				break;
			case GT_Top:
				topnr=t->d.val;
				tch=TRUE;
				break;
			case GT_FrameType:
				ftype=t->d.val;
				break;
			case GT_InnerBorder:
				ibdrw=t->d.val;
				vch=TRUE;
				break;
			case GT_Border:
				bdrw=t->d.val;
				vch=TRUE;
				break;
			case GT_Font:
			case GT_FontList:
				fch=TRUE;
				break;
			case GT_Foreground:
				cch=TRUE;
				break;
		}
	}
	if(tch)
	{
		if(topnr>totnr-visnr)
			topnr=(totnr>visnr ? totnr-visnr : 0);
		tnr=topnr-tnr;
		if(tnr<0)
		{
			for(tnr=-tnr; tnr; tnr--)
				topi=(GtListItem *)GtDPrev(&ihd,&topi->lnk);
		}
		else
		{
			for(; tnr; tnr--)
				topi=(GtListItem *)GtDNext(&ihd,&topi->lnk);
		}
	}
	if(mch)
	{
		if((mode==GTLM_READ) && seli)
		{
			seli->sel=FALSE;
			cd.item=seli;
			cd.sel=seli->sel;
			cd.multi=0;
			sel.CallCB(this,&cd);
			seli=NULL;
		}
		else if(mode==GTLM_SINGLE)
		{
			for(l=(GtListItem *)GtDFirst(&ihd); l;
					l=(GtListItem *)GtDNext(&ihd,&l->lnk))
			{
				if((l!=seli) && (l->sel))
				{
					l->sel=FALSE;
					cd.item=l;
					cd.sel=l->sel;
					cd.multi=0;
					sel.CallCB(this,&cd);
				}
			}
		}
	}
	if(cch)
	{
		for(l=(GtListItem *)GtDFirst(&ihd); l;
				l=(GtListItem *)GtDNext(&ihd,&l->lnk))
		{
			if(l->color==col)
				l->color=gfx->cols[GTFORE];
		}
	}
	if(fch || vch)
	{
		ibw=ibdrw+bdrw+(space ? 1 : 0);
		gtfnt->MaxExtents(&xcs);
		foff=xcs.ascent;
		fh=foff+xcs.descent;
		this->SetMinMaxSize(GTMAX(xcs.width+2*ibw,2*ibw+1),GTMAX(fh+2*ibw,2*ibw+1),
												GTS_MAX,GTS_MAX);
	}
	if(vch || mch || cch || tch)
	{
		if(tch && !vch && !mch && !cch && gfx)
		{
			gtwin->Clear(ibw-1,ibw-1,xsize-2*ibw+2,ysize-2*ibw+2);
			this->Draw(NULL);
		}
		else
			this->Refresh();
	}
	return(TRUE);
}

/************************************************************************\
*
* Get list attributes.
*
\************************************************************************/
void GtList::Get(GtTag *tgs)
{
	GtTag *t;
	long *ptr;
	
	GtWidget::Get(tgs);
	while((t=GtNextTag(&tgs)))
	{
		if(!(ptr=(long *)t->d.ptr))
			continue;
		switch(t->tag)
		{
			case GT_Mode:
				*ptr=mode;
				break;
			case GT_Spacing:
				*ptr=space;
				break;
			case GT_Top:
				*ptr=topnr;
				break;
			case GT_VisibleItems:
				*ptr=visnr;
				break;
			case GT_TotalItems:
				*ptr=totnr;
				break;
			case GT_FrameType:
				*ptr=ftype;
				break;
			case GT_InnerBorder:
				*ptr=ibdrw;
				break;
			case GT_Border:
				*ptr=bdrw;
				break;
		}
	}
}

/************************************************************************\
*
* Check if class is (sub)class of named class.
*
\************************************************************************/
sbool GtList::IsSubClassOf(char *cn)
{
	if(strcmp(cn,"GtList")==0)
		return(TRUE);
	else
		return(GtWidget::IsSubClassOf(cn));
}

/************************************************************************\
*
* Set list event mask.
*
\************************************************************************/
void GtList::SetEMask(ulong em)
{
	GtWidget::SetEMask(em|ExposureMask|ButtonPressMask|ButtonReleaseMask|
												Button1MotionMask);
}

/************************************************************************\
*
* Add list callback.
*
\************************************************************************/
void GtList::AddCB(ulong tag,GtCallB cb,GtUDPtr ud)
{
	switch(tag)
	{
		case GT_Select:
			sel.AddCB(cb,ud);
			break;
		case GT_ValueChange:
			chg.AddCB(cb,ud);
			break;
		default:
			GtWidget::AddCB(tag,cb,ud);
			break;
	}
}

/************************************************************************\
*
* Remove list callback.
*
\************************************************************************/
void GtList::RemCB(ulong tag,GtCallB cb)
{
	switch(tag)
	{
		case GT_Select:
			sel.RemCB(cb);
			break;
		case GT_ValueChange:
			chg.RemCB(cb);
			break;
		default:
			GtWidget::RemCB(tag,cb);
			break;
	}
}

/************************************************************************\
*
* Clear list from items.
*
\************************************************************************/
void GtList::ClrListItems(void)
{
	GtListChangeCD cd;
	GtListItem *l;
	
	while((l=(GtListItem *)GtDRemHead(&ihd)))
		free(l);
	topi=NULL;
	topnr=0;
	seli=NULL;
	totnr=0;
	cd.top=topnr;
	cd.visible=visnr;
	cd.total=totnr;
	chg.CallCB(this,&cd);
	this->Refresh();
}

/************************************************************************\
*
* Add items from user list header.
*
\************************************************************************/
void GtList::AddItems(GtDLink *lhd,ulong pos)
{
	GtListChangeCD cd;
	GtListItem *l,*t;
	ulong p=0;

	if(!lhd)
		return;
	p=GTMIN(pos,totnr);
	for(l=(GtListItem *)GtDFirst(&ihd); l && pos;
			l=(GtListItem *)GtDNext(&ihd,&l->lnk))
	{
		pos--;
	}
	if(l)
	{
		while((t=(GtListItem *)GtDRemTail(lhd)))
		{
			if(t->str)
				t->len=strlen(t->str);
			t->sel=FALSE;
			GtDAddBefore(&l->lnk,&t->lnk);
			totnr++;
		}
	}
	else
	{
		while((t=(GtListItem *)GtDRemHead(lhd)))
		{
			if(t->str)
				t->len=strlen(t->str);
			t->sel=FALSE;
			GtDAddTail(&ihd,&t->lnk);
			totnr++;
		}
	}
	if(topi)
	{
		topnr=0;
		for(l=(GtListItem *)GtDFirst(&ihd); l;
				l=(GtListItem *)GtDNext(&ihd,&l->lnk))
		{
			if(l==topi)
				break;
			topnr++;
		}
	}
	else
	{
		topi=(GtListItem *)GtDFirst(&ihd);
		topnr=0;
	}
	cd.top=topnr;
	cd.visible=visnr;
	cd.total=totnr;
	chg.CallCB(this,&cd);
	if((p>=topnr) && (p<topnr+visnr))
		this->Refresh();
}

/************************************************************************\
*
* Add user item.
*
\************************************************************************/
sbool GtList::AddItem(GtListItem *l,ulong pos)
{
	GtDLink hd;

	if(!l)
		return(FALSE);
	GtDInit(&hd);
	GtDAddTail(&hd,&l->lnk);
	this->AddItems(&hd,pos);
	return(TRUE);
}

/************************************************************************\
*
* Add list allocated item.
*
\************************************************************************/
sbool GtList::AddItem(char *str,ulong pos)
{
	GtListItem *l;

	if((l=(GtListItem *)malloc(sizeof(GtListItem))))
	{
		l->str=str;
		l->color=gfx->cols[GTFORE];
		this->AddItem(l,pos);
		return(TRUE);
	}
	else
	{
		GtError(this,GTEC_NOMEMORY,"No memory for list item!\n");
		return(FALSE);
	}
}

/************************************************************************\
*
* Add list allocated item and set color.
*
\************************************************************************/
sbool GtList::AddItem(char *str,ulong pos,ulong col)
{
	GtListItem *l;
	
	if((l=(GtListItem *)malloc(sizeof(GtListItem))))
	{
		l->str=str;
		l->color=col;
		return(this->AddItem(l,pos));
	}
	else
	{
		GtError(this,GTEC_NOMEMORY,"No memory for list item!\n");
		return(FALSE);
	}
}

/************************************************************************\
*
* Remove item from list.
*
\************************************************************************/
void GtList::RemItem(GtListItem *l)
{
	GtListChangeCD cd;
	GtListItem *t;
	ulong p;
	
	if(!l)
		return;
	if(l==topi)
	{
		if((t=(GtListItem *)GtDPrev(&ihd,&topi->lnk)))
		{
			topnr--;
			topi=t;
		}
		else
		{
			topi=(GtListItem *)GtDFirst(&ihd);
			topnr=0;
		}
		p=topnr;
		GtDRemove(&l->lnk);
		totnr--;
	}
	else
	{
		p=0;
		for(t=(GtListItem *)GtDFirst(&ihd); t;
				t=(GtListItem *)GtDNext(&ihd,&t->lnk))
		{
			if(t==l)
				break;
			p++;
		}
		if(t)
		{
			GtDRemove(&l->lnk);
			totnr--;
		}
	}
	cd.top=topnr;
	cd.visible=visnr;
	cd.total=totnr;
	chg.CallCB(this,&cd);
	if((p>=topnr) && (p<topnr+visnr))
		this->Refresh();
}

/************************************************************************\
*
* Remove item at given position from list.
*
\************************************************************************/
void GtList::RemItem(ulong pos)
{
	GtListItem *l;
	
	for(l=(GtListItem *)GtDFirst(&ihd); l && pos;
			l=(GtListItem *)GtDNext(&ihd,&l->lnk))
	{
		pos--;
	}
	this->RemItem(l);
}

/************************************************************************\
*
* Remove item with given string from list.
*
\************************************************************************/
void GtList::RemItem(char *str)
{
	GtListItem *l;
	
	for(l=(GtListItem *)GtDFirst(&ihd); l;
			l=(GtListItem *)GtDNext(&ihd,&l->lnk))
	{
		if(l->str && (strcmp(l->str,str)==0))
			break;
	}
	this->RemItem(l);
}

/************************************************************************\
*
* Select an item in list.
*
\************************************************************************/
void GtList::SelectItem(GtListItem *l)
{
	GtListSelectCD cd;
	GtListItem *t;
	ulong p=0;
	
	if(!l)
		return;
	for(t=(GtListItem *)GtDFirst(&ihd); t;
			t=(GtListItem *)GtDNext(&ihd,&t->lnk))
	{
		if(t==l)
			break;
		p++;
	}
	if(t)
	{
		if((mode!=GTLM_MULTI) && seli)
		{
			seli->sel=!seli->sel;
			cd.item=seli;
			cd.sel=seli->sel;
			cd.multi=0;
			sel.CallCB(this,&cd);
		}
		if(mode!=GTLM_READ)
		{
			seli=t;
			seli->sel=!seli->sel;
			cd.item=seli;
			cd.sel=seli->sel;
			cd.multi=0;
			sel.CallCB(this,&cd);
		}
		if((p>=topnr) && (p<topnr+visnr))
			this->Refresh();
	}
}

/************************************************************************\
*
* Select item at a given position.
*
\************************************************************************/
void GtList::SelectItem(ulong pos)
{
	GtListItem *l;
	
	for(l=(GtListItem *)GtDFirst(&ihd); l && pos;
			l=(GtListItem *)GtDNext(&ihd,&l->lnk))
	{
		pos--;
	}
	this->SelectItem(l);
}

/************************************************************************\
*
* Select item with a given string.
*
\************************************************************************/
void GtList::SelectItem(char *str)
{
	GtListItem *l;
	
	for(l=(GtListItem *)GtDFirst(&ihd); l;
			l=(GtListItem *)GtDNext(&ihd,&l->lnk))
	{
		if(l->str && (strcmp(l->str,str)==0))
			break;
	}
	this->SelectItem(l);
}

/************************************************************************\
*
* Get next (or first) selected item after given item.
*
\************************************************************************/
GtListItem *GtList::GetSelected(GtListItem *l)
{
	if(!l)
	{
		l=(GtListItem *)GtDFirst(&ihd);
		if(l && l->sel)
			return(l);
	}
	for(l=(GtListItem *)GtDNext(&ihd,&l->lnk); l; 
			l=(GtListItem *)GtDNext(&ihd,&l->lnk))
	{
		if(l->sel)
			return(l);
	}
	return(NULL);
}

/************************************************************************\
*
* Return next/previous list item or first/last list item.
*
\************************************************************************/
GtListItem *GtList::GetNextPrevItem(GtListItem *l,sbool next)
{
	if(l)
	{
		if(next)
			return((GtListItem *)GtDNext(&ihd,&l->lnk));
		else
			return((GtListItem *)GtDPrev(&ihd,&l->lnk));
	}
	else
	{
		if(next)
			return((GtListItem *)GtDFirst(&ihd));
		else
			return((GtListItem *)GtDLast(&ihd));
	}
}

/************************************************************************\
*
* Show a item by minimum scrolling.
*
\************************************************************************/
void GtList::ShowItem(GtListItem *l)
{
	GtListChangeCD cd;
	GtListItem *tmp;
	ulong pos=0,n;
	
	for(tmp=(GtListItem *)GtDFirst(&ihd); tmp; 
			tmp=(GtListItem *)GtDNext(&ihd,&tmp->lnk))
	{
		if(tmp==l)
			break;
		pos++;
	}
	if(tmp==l)
	{
		if(pos<topnr)
		{
			for(n=topnr-pos; n; n--)
				topi=(GtListItem *)GtDPrev(&ihd,&topi->lnk);
			topnr=pos;
			cd.top=topnr;
			cd.visible=visnr;
			cd.total=totnr;
			chg.CallCB(this,&cd);
			this->Refresh();
		}
		else if(pos>topnr+visnr-1)
		{
			for(n=pos-topnr-visnr+1; n; n--)
				topi=(GtListItem *)GtDNext(&ihd,&topi->lnk);
			topnr=pos-visnr+1;
			cd.top=topnr;
			cd.visible=visnr;
			cd.total=totnr;
			chg.CallCB(this,&cd);
			this->Refresh();
		}
	}
}
				
/************************************************************************\
*
* Find position of item.
*
\************************************************************************/
ulong GtList::FindItemPos(GtListItem *l)
{
	GtListItem *tmp;
	ulong pos=0;
	
	for(tmp=(GtListItem *)GtDFirst(&ihd); tmp; 
			tmp=(GtListItem *)GtDNext(&ihd,&tmp->lnk))
	{
		if(tmp==l)
			return(pos);
		pos++;
	}
	return(GTP_LAST);
}
	
/************************************************************************\
*
* Change list size.
*
\************************************************************************/
void GtList::SetPosSize(word xp,word yp,word xs,word ys)
{
	GtListChangeCD cd;
	long tnr;
	ulong tp,tv,tt;
	
	tp=topnr;
	tv=visnr;
	tt=totnr;
	GtWidget::SetPosSize(xp,yp,xs,ys);
	rect.x=ibw;
	rect.y=ibw;
	rect.width=xsize-2*ibw;
	rect.height=ysize-2*ibw;
	if(gfx)
		gtfore->SetClipRectangles(0,0,&rect,1,YXBanded);
	visnr=(ysize-2*ibw+space)/(fh+space);
	tnr=topnr;
	if(totnr-topnr<visnr)
		topnr=(totnr>visnr ? totnr-visnr : 0);
	tnr=topnr-tnr;
	if(tnr<0)
	{
		for(tnr=-tnr; tnr; tnr--)
			topi=(GtListItem *)GtDPrev(&ihd,&topi->lnk);
	}
	else
	{
		for(; tnr; tnr--)
			topi=(GtListItem *)GtDNext(&ihd,&topi->lnk);
	}
	if((topnr!=tp) || (visnr!=tv) || (totnr!=tt))
	{
		cd.top=topnr;
		cd.visible=visnr;
		cd.total=totnr;
		chg.CallCB(this,&cd);
	}
}

/************************************************************************\
*
* Draw list.
*
\************************************************************************/
sbool GtList::Draw(XEvent *xe)
{
	ulong n;
	int yp;
	GtListItem *l;
	ulong col;
	
	if(xe && ((xe->type!=Expose) || (xe->xexpose.count!=0)))
		return(TRUE);
	n=0;
	yp=ibw+space/2;
	col=gfx->cols[GTFORE];
	for(l=topi; l && (n<visnr);
			l=(GtListItem *)GtDNext(&ihd,&l->lnk))
	{
		if(l->sel)
		{
			gtfill->DrawRectangle(gtwin,ibw,yp,xsize-2*ibw,fh-1,TRUE);
			if(space)
			{
				GT3DBOX(gtlht,gtdrk,gtwin,ibw-1,yp-1,xsize-2*ibw+2,fh+1);
			}
		}
		if(l->str)
		{
			if(col!=l->color)
			{
				gtfore->SetForeground(l->color);
				col=l->color;
			}
			gtfore->DrawString(gtwin,ibw,yp+foff,l->str,l->len,FALSE);
		}
		yp+=fh+space;
		n++;
	}
	if(col!=gfx->cols[GTFORE])
		gtfore->SetForeground(gfx->cols[GTFORE]);
	if(flags & GTWF_DISABLED)
		gtdis->DrawRectangle(gtwin,0,0,xsize,ysize,TRUE);
	if(((mode==GTLM_READ) && (ftype>GTF_LAST_NR)) || (ftype==GTF_IN) ||
		(ftype==GTF_DBL_IN))
	{
		for(n=0; n<bdrw; n++)
		{
			GT3DBOX(gtdrk,gtlht,gtwin,n,n,xsize-2*n,ysize-2*n);
		}
	}
	else
	{
		for(n=0; n<bdrw; n++)
		{
			GT3DBOX(gtlht,gtdrk,gtwin,n,n,xsize-2*n,ysize-2*n);
		}
	}
	return(TRUE);
}

/************************************************************************\
*
* Mouse list events.
*
\************************************************************************/
sbool GtList::Mouse(XEvent *xe)
{
	GtListSelectCD cd;
	GtListItem *tmp;
	ulong n;
	int xp,yp;
	
	switch(xe->type)
	{
		case ButtonPress:
			if(flags & GTWF_DISABLED)
				break;
			if(xe->xbutton.button==Button1)
			{
				lflags|=GTLF_PRESS;
				xp=xe->xbutton.x;
				yp=xe->xbutton.y;
				if((xp<ibw) || (xp>=xsize-ibw) || (yp<ibw) || (yp>=ysize-ibw))
					return(TRUE);
				// Find list node under pointer
				cy=ibw+space/2;
				n=0;
				for(curri=(GtListItem *)topi; curri && (n<visnr);
						curri=(GtListItem *)GtDNext(&ihd,&curri->lnk))
				{
					if((yp>=cy) && (yp<cy+fh+space))
						break;
					cy+=fh+space;
					n++;
				}
				// Check if any node is under pointer
				if(curri && (n<visnr))
				{
					// Highlight node under pointer
					gtlht->DrawRectangle(gtwin,ibw,cy,xsize-2*ibw,fh-1,TRUE);
					if(space)
					{
						GT3DBOX(gtlht,gtdrk,gtwin,ibw-1,cy-1,xsize-2*ibw+2,fh+1);
					}
					if(curri->str)
					{
						if(curri->color!=gfx->cols[GTFORE])
							gtfore->SetForeground(curri->color);
						gtfore->DrawString(gtwin,ibw,cy+foff,curri->str,curri->len,FALSE);
						if(curri->color!=gfx->cols[GTFORE])
							gtfore->SetForeground(gfx->cols[GTFORE]);
					}
				}
				else
					curri=NULL;
				return(TRUE);
			}
			break;
		case ButtonRelease:
			if((lflags & GTLF_PRESS) && (xe->xbutton.button==Button1))
			{
				lflags&=~GTLF_PRESS;
				xp=xe->xbutton.x;
				yp=xe->xbutton.y;
				if((xp<ibw) || (xp>=xsize-ibw) || (yp<ibw) || (yp>=ysize-ibw))
					return(TRUE);
				// Check if any node is under pointer
				if(curri)
				{
					// Check if double clicked on node
					if(xe->xbutton.time-last>(ulong)dsp->GetMCTime())
						mclk=1;
					else
						mclk++;
					last=xe->xbutton.time;
					// Reset multiclick if different nodes clicked
					if(curri!=seli)
						mclk=1;
					// Toggle on/off only when no multiclick
					if(mclk==1)
					{
						// Stop toggle off when mode is GTLM_ALWAYSONE
						if((mode==GTLM_ALWAYSONE) && !curri->sel)
							curri->sel=!curri->sel;
					}
					// Redraw node under pointer
					if(!curri->sel)
					{
						// Node is off
						if(space)
							gtwin->Clear(ibw-1,cy-1,xsize-2*ibw+2,fh+1);
						else
							gtwin->Clear(ibw,cy,xsize-2*ibw,fh-1);
					}
					else
					{
						// Node is on
						gtfill->DrawRectangle(gtwin,ibw,cy,xsize-2*ibw,fh-1,TRUE);
						if(space)
						{
							GT3DBOX(gtlht,gtdrk,gtwin,ibw-1,cy-1,xsize-2*ibw+2,fh+1);
						}
					}
					if(curri->str)
					{
						if(curri->color!=gfx->cols[GTFORE])
							gtfore->SetForeground(curri->color);
						gtfore->DrawString(gtwin,ibw,cy+foff,curri->str,curri->len,FALSE);
						if(curri->color!=gfx->cols[GTFORE])
							gtfore->SetForeground(gfx->cols[GTFORE]);
					}
					// Check if previous selected node has to be toggled off
					if((mode!=GTLM_MULTI) && seli && (seli!=curri))
					{
						// Find previous selected node if it is visible
						n=0;
						yp=ibw+space/2;
						for(tmp=topi; tmp && (n<visnr);
								tmp=(GtListItem *)GtDNext(&ihd,&tmp->lnk))
						{
							if((tmp==seli) && (n<visnr))
							{
								// Node is visible so lets redraw it unmarked
								if(space)
									gtwin->Clear(ibw-1,yp-1,xsize-2*ibw+2,fh+1);
								else
									gtwin->Clear(ibw,yp,xsize-2*ibw,fh-1);
								if(tmp->str)
								{
									if(tmp->color!=gfx->cols[GTFORE])
										gtfore->SetForeground(tmp->color);
									gtfore->DrawString(gtwin,ibw,yp+foff,tmp->str,tmp->len,FALSE);
									if(tmp->color!=gfx->cols[GTFORE])
										gtfore->SetForeground(gfx->cols[GTFORE]);
								}						
							}
							yp+=fh+space;
						}
						// Toggle node off and call callback
						seli->sel=FALSE;
						cd.item=seli;
						cd.sel=seli->sel;
						cd.multi=0;
						sel.CallCB(this,&cd);
					}
					// Set new selected node and call callback
					seli=curri;
					cd.item=curri;
					cd.multi=mclk;
					cd.sel=curri->sel;
					sel.CallCB(this,&cd);
				}
				return(TRUE);
			}
			break;
		case MotionNotify:
			if(lflags & GTLF_PRESS)
			{
				xp=xe->xbutton.x;
				yp=xe->xbutton.y;
				if((xp<ibw) || (xp>=xsize-ibw) || (yp<ibw) || (yp>=ysize-ibw))
				{
					// Check if any node is under pointer
					if(curri)
					{
						// Node is under pointer and it has to be redrawn unhighlighted
						if(!curri->sel)
						{
							// Node is off
							if(space)
								gtwin->Clear(ibw-1,cy-1,xsize-2*ibw+2,fh+1);
							else
								gtwin->Clear(ibw,cy,xsize-2*ibw,fh-1);
						}
						else
						{
							// Node is on
							gtfill->DrawRectangle(gtwin,ibw,cy,xsize-2*ibw,fh-1,TRUE);
							if(space)
							{
								GT3DBOX(gtlht,gtdrk,gtwin,ibw-1,cy-1,xsize-2*ibw+2,fh+1);
							}
						}
						if(curri->str)
						{
							if(curri->color!=gfx->cols[GTFORE])
								gtfore->SetForeground(curri->color);
							gtfore->DrawString(gtwin,ibw,cy+foff,curri->str,curri->len,FALSE);
							if(curri->color!=gfx->cols[GTFORE])
								gtfore->SetForeground(gfx->cols[GTFORE]);
						}
						curri=NULL;
					}
					return(TRUE);
				}
				if(curri)
				{
					if((yp>=cy) && (yp<cy+fh))
						return(TRUE);
					// Remove highlight from node under pointer
					if(!curri->sel)
					{
						// Node is off
						if(space)
							gtwin->Clear(ibw-1,cy-1,xsize-2*ibw+2,fh+1);
						else
							gtwin->Clear(ibw,cy,xsize-2*ibw,fh-1);
					}
					else
					{
						// Node is on
						gtfill->DrawRectangle(gtwin,ibw,cy,xsize-2*ibw,fh-1,TRUE);
						if(space)
						{
							GT3DBOX(gtlht,gtdrk,gtwin,ibw-1,cy-1,xsize-2*ibw+2,fh+1);
						}
					}
					if(curri->str)
					{
						if(curri->color!=gfx->cols[GTFORE])
							gtfore->SetForeground(curri->color);
						gtfore->DrawString(gtwin,ibw,cy+foff,curri->str,curri->len,FALSE);
						if(curri->color!=gfx->cols[GTFORE])
							gtfore->SetForeground(gfx->cols[GTFORE]);
					}
				}
				// Find new node under pointer
				cy=ibw+space/2;
				n=0;
				for(curri=(GtListItem *)topi; curri && (n<visnr);
						curri=(GtListItem *)GtDNext(&ihd,&curri->lnk))
				{
					if((yp>=cy) && (yp<cy+fh+space))
						break;
					cy+=fh+space;
					n++;
				}
				// Check if any node is under pointer
				if(curri && (n<visnr))
				{
					// Node is under pointer so highlight it
					gtlht->DrawRectangle(gtwin,ibw,cy,xsize-2*ibw,fh-1,TRUE);
					if(space)
					{
						GT3DBOX(gtlht,gtdrk,gtwin,ibw-1,cy-1,xsize-2*ibw+2,fh+1);
					}
					if(curri->str)
					{
						if(curri->color!=gfx->cols[GTFORE])
							gtfore->SetForeground(curri->color);
						gtfore->DrawString(gtwin,ibw,cy+foff,curri->str,curri->len,FALSE);
						if(curri->color!=gfx->cols[GTFORE])
							gtfore->SetForeground(gfx->cols[GTFORE]);
					}
				}
				else
					curri=NULL;
				return(TRUE);
			}
			break;
	}
	return(FALSE);
}
